// ①WebServerアクセス処理
void setupWebServer() {
  // ②WebServerアクセス処理設定[ホーム画面表示]
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
      sendTop(request);
      Serial.println ( "WEB TOP" );
  });
  // ③WebServerアクセス処理設定[LED制御]
  server.on("/led", HTTP_GET, [](AsyncWebServerRequest *request){
      String setLedVal = request->getParam("o")->value();
      if (setLedVal == "ON") {
        digitalWrite ( LED_PIN, true );
        Serial.println ( "led on" );
      } else {
        digitalWrite ( LED_PIN, false );
        Serial.println ( "led off" );
      }
      request->send(200, "text/plain", "OK");
  });
  // ④WebServerアクセス処理設定[補正処理]  
  server.on("/adjusta", HTTP_GET, [](AsyncWebServerRequest *request){
      if(request->hasParam("n")) {
        adjustAngle = request->getParam("n")->value().toInt();
      }
      if(request->hasParam("r")) {
        int setVal = request->getParam("r")->value().toInt();
        if (0 <= setVal && setVal <= 255 ) {
          righ_on = setVal;
        }
      }
      if(request->hasParam("l")) {
        int setVal = request->getParam("l")->value().toInt();
        if (0 <= setVal && setVal <= 255 ) {
          left_on = setVal;
        }
      }
      // ⑤EEPROMに保存
      st_adjust set_eep_adj;  // ⑤変数を宣言（メモリ上に実体を作成）
      set_eep_adj.setFlag[0]  = 'S';
      set_eep_adj.setFlag[1]  = 'T';
      set_eep_adj.angle_a = adjustAngle;
      set_eep_adj.right_p = righ_on;
      set_eep_adj.left_p  = left_on;
      EEPROM.put<st_adjust>(0, set_eep_adj);  // ⑨EEPROMに書き込み
      EEPROM.commit();
      Serial.println ( "adjest n:" + String(adjustAngle) + " r:" + String(righ_on) + " l:" + String(left_on)  );
      request->send(200, "text/plain", "OK");
  });
  // ⑥WebServerアクセス処理設定[補正データ送信処理] 
  server.on("/get_adj", HTTP_GET, [](AsyncWebServerRequest *request){
      Serial.println ( "get adjust" );
      // ⑦距離測定と送信メッセージ作成
      String sendMsg = "{\"n\":" + String(adjustAngle) + ",\"r\":\"" + String(righ_on) + "\",\"l\":\"" + String(left_on) + "\"}";;
      request->send(200, "application/json", sendMsg);
  });
  // ⑧WebServerアクセス処理設定[超音波センサー角度制御]  
  server.on("/s_angle", HTTP_GET, [](AsyncWebServerRequest *request){
      double angleNum = double(request->getParam("n")->value().toInt());
      Serial.println ( "servo angle:" + String(angleNum) );
      // ⑨現在の角度（アングル）から変更があるか判定
      if (preAngle != int(angleNum) || adjustAngle != preAdjustAngle) {
        // ⑩サーボモータ解説
        // パルス幅(ON時間)で角度が決まる（0度:0.5msec[=>26], 180度:2.4msec[=>123]）
        // 90度の場合) 180:1.9(2.4-0.5) = 90:x => x = 1.9*90/180 = 0.95
        // 0.95+0.5 = 1.45msec => 1.45 / 20(10kHz) *1024 => 75 
        // y度の場合) x = ( (1.9*y/180) + 0.5 ) / 20 * 1024
        int angleTime = ceil( ( (1.9*angleNum/180) + 0.5 ) / 20 * 1024 );
        angleTime = angleTime + adjustAngle;
        // 0度:0.5msec[=>26], 180度:2.4msec[=>123]なので、21から128までで制御
        if (angleTime < 21) {
          angleTime = 21;
        } else if (angleTime > 128) {
          angleTime = 128;          
        }
        // ⑪サーボモータ回転
        ledcWrite(SERV_CH, angleTime);
        Serial.println ( "servo angle control:" + request->getParam("n")->value() + ":" + String(angleTime) );
        // ⑫角度を変更する時間を取って返答
        delay ( 1000 );
      // ⑬変更がないときはサーボモータ処理しない
      } else {
        Serial.println ( "servo angle noChange" );
      }
      // ⑭現在の角度を保持
      preAngle = int(angleNum);
      preAdjustAngle = adjustAngle;
      request->send(200, "text/plain", "OK");
  });
  // ⑮WebServerアクセス処理設定[超音波センサー距離測定]  
  server.on("/update_d", HTTP_GET, [](AsyncWebServerRequest *request){
      Serial.println ( "update distance" );
      // ⑯距離測定と送信メッセージ作成
      String sendMsg = updateSensor();
      request->send(200, "application/json", sendMsg);
  });
  // ⑰WebServerアクセス処理設定[Car前進、停止、後退制御]   
  server.on("/cargo", HTTP_GET, [](AsyncWebServerRequest *request){
      // ⑱要求内容を判断しカー制御
      if ( request->getParam("a")->value() == "STOP" ){
         Serial.println("STOP");
         car_stop();
      } else if ( request->getParam("a")->value() == "BACK" ){
         Serial.println("BACK");
         car_back();
      } else {
        Serial.println("GO");
        car_go();
      }
      request->send(200, "text/plain", "OK");
  });
  // ⑲WebServerアクセス処理設定[Car右左折制御]   
  server.on("/carturn", HTTP_GET, [](AsyncWebServerRequest *request){
      // ⑳要求内容を判断しカー制御
      if ( request->getParam("a")->value() == "RIGHT" ){
        Serial.println("RIGHT");
        car_right ( request->getParam("n")->value().toInt() );
      } else {
        Serial.println("LEFT");
        car_left ( request->getParam("n")->value().toInt() );
      }
      request->send(200, "text/plain", "OK");
  });
  // ㉑Responseヘッダオプション追加処理  
  DefaultHeaders::Instance().addHeader("Access-Control-Allow-Origin", "*");
  // ㉒WebServer起動  
  server.begin();
}

// ㉓HTMLトップ返答処理
void sendTop( AsyncWebServerRequest *request ) {
  // ㉔返答HTML定義
  char *htmltop = 
  "<!doctype html><html class='fullscrn'><head><meta charset='UTF-8'/>\r\n"
  "<meta name='viewport' content='width=device-width'/>\r\n"
  "<link rel=\"icon\" href=\"http://kit.socinno.com/wp-content/uploads/2019/09/favicon.ico\">\r\n"
  "<style type='text/css'><!--\r\n"
  "h1 { margin: 13px 0; font-size: 28px; line-height: 70%; }\r\n"
  "#contents { width: 100%; max-width: 320px; }\r\n"
  "footer { text-align: right; }\r\n"
  ".fullscrn { width: 100%; height: 100%; }\r\n"
  ".cosmos { color: white; background-color: #797572;\r\n"
  "background-image: linear-gradient(-173deg, rgba(255,255,255,0.20) 0%, #000000 100%),\r\n"
  "linear-gradient(72deg, rgba(255,255,255,0.25) 25%, rgba(0,0,0,0.25) 100%),\r\n"
  "radial-gradient(47% 102%, rgba(255,255,255,0.50) 0%, rgba(21,24,32,0.60) 120%);background-blend-mode: multiply; }\r\n"
  "--></style>\r\n"
  "<title>IoTスマートカー</title></head>\r\n"
  "<body class='cosmos fullscrn'><center><div id='contents'><header><h3>IoTスマートカー制御画面</h3></header>\r\n"
  "<form id='canform' method='get' action='javascript:void(0);'>\r\n"
  "\r\n"
  "<div id='dcan1' style=\"display:none\"><canvas id=\"ccanvas\" width=\"176\" height=\"144\" style=\"background-color:grey; float:left\"></canvas>\r\n"
  "<input id='chkCan1' type=\"checkbox\" checked onclick=\"canDisp(0);\">映像表示<br>\r\n"
  "<span style='line-height:25px;text-align:center;'><br><button id='videobtn' type='button' style='width:70px;height:25px' onClick=\"wsConnect()\" >開始</button><br>\r\n"
  "<font color=red><span id=\"msg\" style='vertical-align:middle;'>DISCONNECT</span></font><br>\r\n"
  "<button id='endbtn' type='button' style='width:70px;height:25px' onClick=\"wsDisconnect()\" >終了</button></span><p style=\"clear: both;\"></p><br></div>\r\n"
  "\r\n"
  "<table width=100%><tr><td>カー制御</td><td align=right><span id='dcan2' style=\"display:inline\"><input id='chkCan2' type=\"checkbox\" onclick=\"canDisp(1);\">映像表示</span></td></tr></table>\r\n"
  "<table width=100%><tr><td></td><td><button id='gobtn' type='button' style='width:100px;height:50px' onClick='car_cnt(\"GO\")' >GO</button></td><td></td></tr>\r\n"
  "<tr><td><button id='leftbtn' type='button' style='width:100px;height:50px' onClick='car_turn(\"LEFT\")' >LEFT</button></td>\r\n"
  "<td><button id='stopbtn' type='button' style='width:100px;height:50px' onClick='car_cnt(\"STOP\")' >STOP</button></td>\r\n"
  "<td><button id='rightbtn' type='button' style='width:100px;height:50px' onClick='car_turn(\"RIGHT\")' >RIGHT</button></td></tr>\r\n"
  "<tr><td style='vertical-align:top;'>左<input type=\"text\" id=\"leftms\" size=4 value=250>ms</td>\r\n"
  "<td><button id='backbtn' type='button' style='width:100px;height:50px' onClick='car_cnt(\"BACK\")' >BACK</button></td>\r\n"
  "<td style='vertical-align:top; text-align:right;'>右<input type=\"text\" id=\"rightms\" size=4 value=250>ms</td></tr><table><br>\r\n"
  "\r\n"
  "<table width=100%><tr><td>センサー制御</td><td align=right>距離：<span id='distance'>0</span>cm</td></tr></table>\r\n"
  "<table width=100%><tr><td><button id='langbtn' type='button' style='width:100px;height:50px' onClick='sen_cnt(\"LEFT\")' >LEFT</button></td>\r\n"
  "<td><button id='cangbtn' type='button' style='width:100px;height:50px' onClick='sen_cnt(\"CENTER\")' >CENTER</button></td>\r\n"
  "<td><button id='rangbtn' type='button' style='width:100px;height:50px' onClick='sen_cnt(\"RIGHT\")' >RIGHT</button></td></tr>\r\n"
  "<tr><td style='vertical-align:top;'><div style='line-height:0.9em;'>左<input type=\"text\" id=\"leftdo\" size=3 value=180>度<br><br><br>左IR:<span id=\"leftir\">無し</span></div></td>\r\n"
  "<td><button id='measurebtn' type='button' style='width:100px;height:50px' onClick='s_mesure()' >MESUREMENT</button></td>\r\n"
  "<td style='vertical-align:top; text-align:right;'><div style='line-height:0.9em;'>右<input type=\"text\" id=\"rightdo\" size=3 value=0>度<br><br><br>右IR:<span id=\"rightir\">無し</span></div></td></tr><table><br>\r\n"
  "\r\n"
  "<table width=100%><tr><td>バランス調整</td><td align=right><input id='chkadj' type=\"checkbox\" onclick=\"adjDisp();\">設定</td></tr></table>\r\n"
  "<div id='adjtbl' style=\"display:none\">\r\n"
  "<table width=100%><tr><td>角度補正<input type=\"text\" id=\"angleadj\" size=3 value=0></td>\r\n"
  "<td align=right>左出力<input type=\"text\" id=\"leftpow\" size=3 value=255></td>\r\n"
  "<td align=right>右出力<input type=\"text\" id=\"rightpow\" size=3 value=255></td></tr>\r\n"
  "<tr><td></td><td><button id='adjbtn' type='button' style='width:100px;height:30px' onClick='s_adjust()' >ADJUST</button></td><td></td></tr>\r\n"
  "</table></div>\r\n"
  "\r\n"
  "</form>\r\n"
  "\r\n"
  "<div id='dispStatus'></div>\r\n"
  "\r\n"
  "<script language=\"javascript\" type=\"text/javascript\">\r\n"
  "var wsUri;\r\n"
  "var socket = null;\r\n"
  "var tms;\r\n"
  "var msgIn;\r\n"
  "var msg;\r\n"
  "var ctx;\r\n"
  "var width;\r\n"
  "var height;\r\n"
  "var imageData;\r\n"
  "var pixels;\r\n"
  "var fps = 0;\r\n"
  "var jobFlag = false;\r\n"
  "var canFlag = false;\r\n"
  "var wsFlag = false;\r\n"
  "var adjFlag = false;\r\n"
  "window.onload = function(){\r\n"
  " wsUri = \"ws://\" + window.location.hostname + \":81/\";\r\n"
  " msgIn = document.getElementById('msgIn');\r\n"
  " msg = document.getElementById('msg');\r\n"
  " var canv = document.getElementById('ccanvas');\r\n"
  " ctx = canv.getContext('2d');\r\n"
  " width = canv.width;\r\n"
  " height = canv.height;\r\n"
  " imageData = ctx.createImageData( width, 1 );\r\n"
  " pixels = imageData.data;\r\n"
  " var fullpath = location.href\r\n"
  " var path = fullpath.split('?');\r\n"
  " if ( path[1] && path[1] != \"\" ) {\r\n"
  "   var eachData = path[1].split('&');\r\n"
  "   for (var i = 0; i < eachData.length; i++) {\r\n"
  "     var tmpData = eachData[i].split('=');\r\n"
  "     if ( tmpData[0] == \"camera\" && tmpData[1] == \"on\" ) {\r\n"
  "       canFlag = true;\r\n"
  "       canDisp(1);\r\n"
  "     } else if ( tmpData[0] == \"rms\" ) {\r\n"
  "       document.getElementById(\"rightms\").value = tmpData[1];\r\n"
  "     } else if ( tmpData[0] == \"lms\" ) {\r\n"
  "       document.getElementById(\"leftms\").value = tmpData[1];\r\n"
  "     } else if ( tmpData[0] == \"rdo\" ) {\r\n"
  "       document.getElementById(\"rightdo\").value = tmpData[1];\r\n"
  "     } else if ( tmpData[0] == \"ldo\" ) {\r\n"
  "       document.getElementById(\"leftdo\").value = tmpData[1];\r\n"
  "     }\r\n"
  "   }\r\n"
  " }\r\n"
  " if (canFlag) {\r\n"
  "   document.getElementById(\"videobtn\").disabled = true;\r\n"
  "   setTimeout('wsConnect()', 1000);\r\n"
  " }\r\n"
  " get_adj();\r\n"
  "}\r\n"
  "\r\n"
  "function Msg(message){ msg.innerHTML = message;}\r\n"
  "function wsConnect(){\r\n"
  " wsFlag = true;\r\n"
  " document.getElementById(\"videobtn\").disabled = true;\r\n"
  " tms = new Date();\r\n"
  " if(socket == null){\r\n"
  "   socket = new WebSocket(wsUri);\r\n"
  "   socket.binaryType = 'arraybuffer';\r\n"
  "   socket.onopen   = function(evt){ onOpen(evt) };\r\n"
  "   socket.onclose  = function(evt){ onClose(evt) };\r\n"
  "   socket.onmessage = function(evt){ onMessage(evt) };\r\n"
  "   socket.onerror  = function(evt){ onError(evt) };\r\n"
  " }\r\n"
  "}\r\n"
  "function wsDisconnect(){\r\n"
  " wsFlag = false;\r\n"
  " socket.close();\r\n"
  " socket = null;\r\n"
  " Msg('STOP');\r\n"
  " document.getElementById(\"videobtn\").disabled = false;\r\n"
  "}\r\n"
  "function onOpen(evt){ Msg('START');}\r\n"
  "function onClose(evt){\r\n"
  " Msg('DISCONNECT');\r\n"
  " wsClose();\r\n"
  " document.getElementById(\"videobtn\").disabled = false;\r\n"
  "}\r\n"
  "function onError(evt){\r\n"
  " Msg('Err:' + evt.data);\r\n"
  " document.getElementById(\"videobtn\").disabled = false;\r\n"
  "}\r\n"
  "function onMessage(evt){\r\n"
  " var data = evt.data;\r\n"
  " if( typeof data == 'string'){\r\n"
  "   msgIn.innerHTML = data;\r\n"
  " }else if( data instanceof ArrayBuffer){\r\n"
  "   drawLine(evt.data);\r\n"
  " }else if( data instanceof Blob){\r\n"
  "   Msg('Blob data received');\r\n"
  " }\r\n"
  "}\r\n"
  "function wsClose(){\r\n"
  " socket.close();\r\n"
  " socket = null;\r\n"
  " if (wsFlag) {\r\n"
  "   setTimeout('wsConnect()', 1);\r\n"
  " }\r\n"
  "}\r\n"
  "function drawLine(data){\r\n"
  " var buf = new Uint16Array(data);\r\n"
  " for(var y = 0; y < (buf.length)/width; y+=1){\r\n"
  "   var base = 0;\r\n"
  "   for(var x = 0; x < width; x += 1){\r\n"
  "     var c = x + y * width;\r\n"
  "     pixels[base+0] = (buf[c] & 0xf800) >> 8 | (buf[c] & 0xe000) >> 13;\r\n"
  "     pixels[base+1] = (buf[c] & 0x07e0) >> 3 | (buf[c] & 0x0600) >> 9;\r\n"
  "     pixels[base+2] = (buf[c] & 0x001f) << 3 | (buf[c] & 0x001c) >> 2;\r\n"
  "     pixels[base+3] = 255;\r\n"
  "     base += 4;\r\n"
  "   }\r\n"
  "   ctx.putImageData(imageData, 0, y);\r\n"
  " }\r\n"
  " if(y == height) fps+=1;\r\n"
  "}\r\n"
  "function canDisp(chgFlag) {\r\n"
  " if (chgFlag == 0) {\r\n"
  "    document.getElementById(\"dcan1\").style.display=\"none\";\r\n"
  "    document.getElementById(\"dcan2\").style.display=\"inline\";\r\n"
  " } else {\r\n"
  "    document.getElementById(\"dcan1\").style.display=\"block\";\r\n"
  "    document.getElementById(\"dcan2\").style.display=\"none\";\r\n"
  " }\r\n"
  " document.getElementById(\"chkCan1\").checked = true;\r\n"
  " document.getElementById(\"chkCan2\").checked = false;\r\n"
  "}\r\n"
  "function adjDisp() {\r\n"
  " adjFlag = !adjFlag;\r\n"
  " if (adjFlag) {\r\n"
  "    document.getElementById(\"adjtbl\").style.display=\"inline\";\r\n"
  " } else {\r\n"
  "    document.getElementById(\"adjtbl\").style.display=\"none\";\r\n"
  " }\r\n"
  "}\r\n"
  "function car_cnt(cntSig) {\r\n"
  " var url = \"http://\" + window.location.hostname + \"/cargo?a=\" + cntSig;\r\n"
  " htmlacs(url);\r\n"
  "}\r\n"
  "function car_turn(cntSig) {\r\n"
  " var rltime;\r\n"
  " if ( cntSig == \"RIGHT\" ) {\r\n"
  "   rltime = document.getElementById(\"rightms\").value;\r\n"
  " } else {\r\n"
  "   rltime = document.getElementById(\"leftms\").value;\r\n"
  " }\r\n"
  " var url = \"http://\" + window.location.hostname + \"/carturn?a=\" + cntSig + \"&n=\" + rltime;\r\n"
  " htmlacs(url);\r\n"
  "}\r\n"
  "function s_mesure() {\r\n"
  " var url = \"http://\" + window.location.hostname + \"/update_d\";\r\n"
  " htmlacs(url);\r\n"
  "}\r\n"
  "function s_adjust() {\r\n"
  " var adja = document.getElementById(\"angleadj\").value;\r\n"
  " var adjl = document.getElementById(\"leftpow\").value;\r\n"
  " var adjr = document.getElementById(\"rightpow\").value;\r\n"
  " var url = \"http://\" + window.location.hostname + \"/adjusta?n=\" + adja + \"&l=\" + adjl + \"&r=\" + adjr;\r\n"
  " htmlacs(url);\r\n"
  "}\r\n"
  "function get_adj() {\r\n"
  " var url = \"http://\" + window.location.hostname + \"/get_adj\";\r\n"
  " htmlacs(url);\r\n"
  "}\r\n"
  "function sen_cnt(cntSig) {\r\n"
  " var rldo;\r\n"
  " if ( cntSig == \"RIGHT\" ) {\r\n"
  "   rldo = document.getElementById(\"rightdo\").value;\r\n"
  " } else if ( cntSig == \"CENTER\" ) {\r\n"
  "   rldo = 90;\r\n"
  " } else {\r\n"
  "   rldo = document.getElementById(\"leftdo\").value;\r\n"
  " }\r\n"
  " var url = \"http://\" + window.location.hostname + \"/s_angle?n=\" + rldo;\r\n"
  " htmlacs(url);\r\n"
  "}\r\n"
  "function htmlacs(url) {\r\n"
  " if (jobFlag) {\r\n"
  "   document.getElementById('dispStatus').innerHTML = \"<b><font color=red>処理中です。</font></b>\";\r\n"
  " }\r\n"
  " jobFlag = true;\r\n"
  " var xhr = new XMLHttpRequest();\r\n"
  " if(xhr.timeout !== undefined){\r\n"
  "   xhr.ontimeout = function(e){\r\n"
  "     finishJob( \"<b><font color=red>本体への通信に失敗</font></b>しました。\" )\r\n"
  "   };\r\n"
  "   xhr.timeout = 10000;\r\n"
  " }\r\n"
  " xhr.open(\"GET\", url);\r\n"
  " xhr.send(\"\");\r\n"
  " xhr.addEventListener(\"load\",function(ev){\r\n"
  "   var resStr = xhr.responseText;\r\n"
  "   if ( url.indexOf('update_d') != -1 ) {\r\n"
  "     var pRecv =  JSON.parse(resStr);\r\n"
  "     document.getElementById('distance').innerHTML = pRecv.d;\r\n"
  "     if ( pRecv.r == 'L' ) {\r\n"
  "       document.getElementById('rightir').innerHTML = '検出';\r\n"
  "     } else {\r\n"
  "       document.getElementById('rightir').innerHTML = '無し';\r\n"
  "     }\r\n"
  "     if ( pRecv.l == 'L' ) {\r\n"
  "       document.getElementById('leftir').innerHTML = '検出';\r\n"
  "     } else {\r\n"
  "       document.getElementById('leftir').innerHTML = '無し';\r\n"
  "     }\r\n"
  "     finishJob( \"<b><font color=red>正常完了しました。</font></b>\" );\r\n"
  "   } else if ( url.indexOf('get_adj') != -1 ) {\r\n"
  "     var pRecv =  JSON.parse(resStr);\r\n"
  "     document.getElementById('angleadj').value = pRecv.n;\r\n"
  "     document.getElementById('leftpow').value = pRecv.l;\r\n"
  "     document.getElementById('rightpow').value = pRecv.r;\r\n"
  "   } else if ( resStr == \"OK\" ) {\r\n"
  "     finishJob( \"<b><font color=red>正常完了しました。</font></b>\" );\r\n"
  "   } else {\r\n"
  "     finishJob( \"<b><font color=red>処理失敗です。再実行して下さい。</font></b>\" );\r\n"
  "   }\r\n"
  " });\r\n"
  "}\r\n"
  "function finishJob( dispWord ){\r\n"
  " jobFlag = false;\r\n"
  " document.getElementById('dispStatus').innerHTML = dispWord;\r\n"
  "}\r\n"
  "</script>\r\n"
  "<br><footer><font size=-1>Copyright© SOCINNO All rights reserved.</font></footer></div></center></body></html>\r\n"
  ;

  // ㉕返答HTML送信
  request->send(200, "text/html", htmltop);
}
