//*************************************************************************
//  sLab-Car(エスラボ・カー)
//  Scratch3など遠隔制御用【カメラ有り、左右ホイールバランス制御有り】
//*************************************************************************
  // ①ライブラリを読み込み
#include <WiFi.h>
#include <Wire.h>
#include <EEPROM.h>
#include "ESPAsyncWebServer.h"
#include "WebSocketsServer.h"
#include "OV7670.h"

// ####################### ②各種設定(環境設定) #######################
const char *ssid      = "##### SSID #####";
const char *password  = "### PASSWORD ###";
IPAddress ip(192, 168, 1, 123);  // IPアドレス（本機が利用するIP）
IPAddress gateway(192, 168, 1, 1);  // デフォルトゲートウェイ
IPAddress subnet(255, 255, 255, 0); // サブネットマスク
// ###################################################################
// 各種個体バランス設定
int adjustAngle   =   0;    // サーボモータ初期角度補正
byte righ_on      = 255;    // 右モータ用出力[0-255]
byte left_on      = 255;    // 左モータ用出力[0-255]
// ###################################################################


// ③ピン配置
const byte LED_PIN      = 17;  // LED緑
const byte R_WHEEL_IN1  = 13;  // 右モータ制御1
const byte R_WHEEL_IN2  = 12;  // 右モータ制御2
const byte L_WHEEL_IN1  =  2;  // 左モータ制御1
const byte L_WHEEL_IN2  =  4;  // 左モータ制御2
const byte ONPA_TRIG    = 18;  // 超音波トリガー
const byte ONPA_ECHO    =  5;  // 超音波エコー
const byte SERV_PIN     = 27;  // サーボモータ
const byte IR_LEFT_U    = 19;  // 上部、左赤外線センサ
const byte IR_RIGH_U    = 23;  // 上部、右赤外線センサ
const byte SERV_CH      =  7;  // サーバモータ用LEDCチャンネル
const byte RIG1_CH      =  8;  // 右1モータ用LEDCチャンネル
const byte RIG2_CH      =  9;  // 右2モータ用LEDCチャンネル
const byte LEF1_CH      = 10;  // 左1モータ用LEDCチャンネル
const byte LEF2_CH      = 11;  // 左2モータ用LEDCチャンネル
const int  LEDC_HZ      = 490; // モータ用LEDC周波数

// ④カメラピン配置
const camera_config_t cam_conf = {
  .D0 = 36,
  .D1 = 39,
  .D2 = 34,
  .D3 = 35,
  .D4 = 32,
  .D5 = 33,
  .D6 = 25,
  .D7 = 26,
  .XCLK = 15,   // ★27利用時 Wi-Fi接続不可
  .PCLK = 14,
  .VSYNC = 16,
  .xclk_freq_hz = 10000000,     // XCLK 10MHz
  .ledc_timer   = LEDC_TIMER_0,
  .ledc_channel = LEDC_CHANNEL_0 
};

// ⑤カメラ解像度指定
#define CAM_RES     QCIF    // カメラ解像度
#define CAM_WIDTH   176     // カメラ幅
#define CAM_HEIGHT  144     // カメラ高さ
#define CAM_DIV       1     // １画面分割数

// ⑥グローバル変数宣言
// EEPROMで利用する型（構造体）を宣言
struct st_adjust {          // バランス補正値を保存 
  char setFlag[3];          // 'ST'だと設定済み。残り１Byteは予備
  int  angle_a;
  byte  left_p;
  byte right_p;
};

OV7670 cam;                 // camera
AsyncWebServer server(80);  // WebサーバPort80番
WebSocketsServer webSocket = WebSocketsServer(81);
uint8_t clientNum;

bool wsFlag  = false;       // WebSocket状態フラグ(true:接続状態、false:未接続状態)
bool ledFlag = true;        // LED制御フラグ
int dataSize = 2 * CAM_WIDTH * CAM_HEIGHT;      // カメラデータ: 2 * 176 * 144 = 50688 [Byte]
uint8_t wsCamData[50688];   // カメラデータを保持
int preAdjustAngle = 0;     // 前回Angle補正
byte preAngle      = -1;    // 現在の角度を保存し、角度変更があるか把握
byte preCarMove    = 1;     // 現在のCar動作を保存し変更があるか把握　1:初期/Go、2:Left、3:Right、4:Back、5:Other

// ⑦Arduino初期設定処理
void setup() {
  // ⑧処理開始
  Serial.begin(115200);
  Serial.println("sLab-Car Sart"); 
  Wire.begin();
  Wire.setClock(400000);
  // ⑨端子設定
  pinMode ( LED_PIN, OUTPUT );
  pinMode(ONPA_TRIG, OUTPUT);
  pinMode(ONPA_ECHO, INPUT);
  pinMode(IR_LEFT_U, INPUT);
  pinMode(IR_RIGH_U, INPUT);
  ledcSetup(SERV_CH, 50, 10);       // ch:7, Freq:50Hz, 10bit
  ledcAttachPin(SERV_PIN, SERV_CH); // ch:7
  ledcSetup(RIG1_CH, LEDC_HZ, 8);        // ch:8, Freq:490Hz, 8bit
  ledcAttachPin(R_WHEEL_IN1, RIG1_CH); // ch:8
  ledcSetup(RIG2_CH, LEDC_HZ, 8);        // ch:9, Freq:490Hz, 8bit
  ledcAttachPin(R_WHEEL_IN2, RIG2_CH); // ch:9
  ledcSetup(LEF1_CH, LEDC_HZ, 8);        // ch:10, Freq:490Hz, 8bit
  ledcAttachPin(L_WHEEL_IN1, LEF1_CH); // ch:10
  ledcSetup(LEF2_CH, LEDC_HZ, 8);        // ch:11, Freq:490Hz, 8bit
  ledcAttachPin(L_WHEEL_IN2, LEF2_CH); // ch:11
  // ⑩サーボモータ回転(前向き[90度]へ調整) 
  ledcWrite(SERV_PIN, (76 + adjustAngle));
  ledcWrite(RIG1_CH, 0);
  ledcWrite(RIG2_CH, 0);
  ledcWrite(LEF1_CH, 0);
  ledcWrite(LEF2_CH, 0);
  // ⑪EEPROM開始と値取得
  EEPROM.begin(128);  // ④EEPROM開始（128Byte確保）
  st_adjust eep_adj;  // ⑤メモリ上に実体を作成
  EEPROM.get<st_adjust>(0, eep_adj);  // ⑥EEPROMの値を取得
  if (eep_adj.setFlag[0] == 'S' && eep_adj.setFlag[1] == 'T') {
    adjustAngle = eep_adj.angle_a;
    righ_on     = eep_adj.right_p;
    left_on     = eep_adj.left_p;
  }
  Serial.println ( "Start_adj a:" + String(adjustAngle) + " r:" + String(righ_on) + " l:" + String(left_on)  );
  // ⑫無線Wi-Fi接続
  WiFi.config( ip, gateway, subnet );
  WiFi.begin ( ssid, password );
  while ( WiFi.status() != WL_CONNECTED ) { // 接続するまで無限ループ処理
    // ⑬接続中はLEDを１秒毎に点滅する
    ledFlag = !ledFlag;
    digitalWrite(LED_PIN, ledFlag);
    delay ( 1000 );
    Serial.print ( "." );
  }
  // ⑭Wi-Fi接続完了（IPアドレス表示）
  Serial.print ( "Wi-Fi Connected! IP address: " );
  Serial.println ( WiFi.localIP() );
  // ⑮Wi-Fi接続時はLED点灯（Wi-Fi接続状態）
  digitalWrite ( LED_PIN, true );

  // ⑯カメラ初期化処理
  Serial.println( "cam init" );  
  esp_err_t err = cam.init(&cam_conf, CAM_RES, RGB565);
  if(err != ESP_OK){
    Serial.print( "cam.init ERR:" );
    Serial.println(err);
    // ⑰失敗時は15秒後に再起動しRetry
    delay(15000);
    ESP.restart();
  }
  cam.vflip( false );
  // ⑱WebServer起動
  setupWebServer();
  // ⑲WebSocket起動
  webSocket.begin();
  webSocket.onEvent(onWebSocketEvent);
  Serial.println("##### setup completed #####");
}

// ⑳Arduinoループ処理（通常時はこの処理）
void loop(void) {
  // ㉑WebSocket処理
  webSocket.loop();
  // ㉒WebSocket接続判定
  if (wsFlag) {
    // ㉓カメラデータ読み込み処理
    cam.getLines( 1 , wsCamData , CAM_HEIGHT);
    // ㉔映像データ送信処理
    webSocket.sendBIN(clientNum, wsCamData, dataSize);
  }
}
